{
 Unit to handle directing TurboVision screen I/O to a serial
 port. The Borland VIEWS.PAS has been modified to include
 a far pointer (viewremote) to a routine to handle the serial output.
 This could be easily adapted to produce network traffic in the
 same way as serial port traffic. VT100 emulation works fine
 except for recognising <Alt>+key sequences. TeraTerm (freeware)
 (with the META key feature enabled) will send <Alt>+key
 sequences. A serial-to-ethernet adapter would allow for real
 time applications without monitors to run remotely.
}
unit Hl_remot ;
interface

uses Drivers, Objects, Views ;
type
  PSRemote = ^TSRemote ; { serial i/o object }
  TSRemote = object( TView )
               cport : word ; { I/O address of serial port }
               scol  : word ; { current fore/back colours }
               cposx : integer ; { current cursor position X }
               cposy : integer ; { current cursor position Y }
               keystream : string[16] ; { string of incoming keys }
               escaping : boolean ; { TRUE if in an escape sequence }
               Constructor Init( ioport,baud : word );
               Destructor Done ; virtual ;
               Procedure GetComEvent( var event : TEvent ) ; virtual ;
               Procedure Send( toptr,fromptr : pointer ; count : word ) ;
               Function GetChar : word ;
               Procedure PutLine( oline : string ) ;
               Procedure SetColor( icol : byte ) ;
             private
               Procedure InitPort( psb : byte ; baud : word ) ;
               Procedure PutChar( obyte : byte ) ;
               Function GetStat : word ;
               Procedure SetCursor( ilin,icol : integer ) ;
               Procedure UpdateCursor ;
             end ;
var
  sremote : PSRemote ; { the global remote serial object }

const
  RB_600    = $00C0 ;
  RB_2400   = $0030 ; { comport baud rate constants }
  RB_4800   = $0018 ;
  RB_9600   = $000C ;
  RB_19200  = $0006 ;
  RB_38400  = $0003 ;
  RB_57600  = $0002 ;
  RB_115200 = $0001 ;

implementation
uses App, Dialogs, Hl_bits, Hl_strng ;
const
  ESC  = #$1b ;
  KEYCODES : array[ 0..$1F ] of word = { map ctrl character to keycode }
              ( 0,1,2,3,4,5,6,7,kbBack,kbTab,10,11,12,kbEnter,14,15,
                16,17,18,19,20,21,22,23,24,25,26,kbEsc,28,29,30,31 ) ;
  CMAP : array[ 0..7 ] of byte = { maps DOS colours to ANSI colours }
          ( 30,34,32,36,31,35,33,37 ) ;
  COLS = 80 ; { standard application screen size }
  ROWS = 25 ;

{----------------------------------------------------------
 Wrapper around object Send procedure so can be pointed
 to by ViewRemote. Quick and dirty.
----------------------------------------------------------}
Procedure RemoteCall( toptr,fromptr : pointer ; count : word ) ; far ;
begin
  if ( sremote<>NIL ) then
    sremote^.Send( toptr,fromptr,count ) ;
end;

{----------------------------------------------------------
 Initialise serial port and send out terminal reset.
 See InitPort for baud rate constants or RB_xxxx above
----------------------------------------------------------}
Constructor TSRemote.Init( ioport,baud : word ) ;
var
  r : TRect ;
begin
  r.Assign( 0,0,1,1 ) ;
  inherited Init( r ) ;
  cport := ioport ;
  GetStat ;              { read status registers to clear stuff }
  InitPort( 3,baud ) ;   { baud,n,8,1 }
  PutLine( ESC+'[2J' ) ; { clear the remote screen }
  ViewRemote := RemoteCall ; { assign the VIEWS.PAS pointer }
end ;

{----------------------------------------------------------
 Clear the terminal screen before quitting
----------------------------------------------------------}
Destructor TSRemote.Done ;
begin
  inherited Done ;
  SetColor( $F0 ) ; { black text on white background }
  PutLine( ESC+'[2J' ) ; { clear the remote screen }
end ;

{----------------------------------------------------------
 Get a serial input stream from the terminal and convert
 to keyevents. Alt-<key> shows up as <Esc>key in TeraTerm.
 Update cursor position if required. This is usually called
 from idle.
----------------------------------------------------------}
Procedure TSRemote.GetComEvent( var event : TEvent ) ;
var
  ichar : word ;
begin
  UpdateCursor ;
  event.what := evNothing ;
  ichar := GetChar ;
  while (Hi(ichar) and $01)<>0 do
    begin
      if escaping then
        begin { currently processing escape code }
          keystream := keystream+Char(Lo(ichar)) ;
          case keystream[1] of
            '[','O' : { <esc>[ or <esc>O for cursor control and PF1..PF4 }
                  begin
                    if Length( keystream )>1 then
                      begin
                        event.what := evKeydown ;
                        case keystream[2] of
                          'A' : event.keycode := kbUp ;
                          'B' : event.keycode := kbDown ;
                          'C' : event.keycode := kbRight ;
                          'D' : event.keycode := kbLeft ;
                          'P' : event.keycode := kbF1 ;
                          'Q' : event.keycode := kbF2 ;
                          'R' : event.keycode := kbF3 ;
                          'S' : event.keycode := kbF4 ;
                        end ;
                        keystream := '' ;
                        escaping := FALSE ;
                      end ;
                  end ;
          else
            begin { second character is not [ or O }
              case Char(Lo(ichar)) of
                'a'..'z' : begin { esc+key same as <Alt>-key for TeraTerm }
                             event.what := evKeydown ;
                             event.keycode := GetAltCode(Char(Lo(ichar)));
                             escaping := FALSE ;
                             keystream := '' ;
                           end ;
                    #$1B : begin {ESC+ESC as escape}
                             event.what := evKeydown ;
                             event.keycode := kbEsc ;
                             escaping := FALSE ;
                             keystream := '' ;
                           end ;
                'A'..'D' : begin { messed up cursor key so try to recover }
                             event.what := evKeydown ;
                             case keystream[1] of
                               'A' : event.keycode := kbUp ;
                               'B' : event.keycode := kbDown ;
                               'C' : event.keycode := kbRight ;
                               'D' : event.keycode := kbLeft ;
                             else
                               event.what := evNothing ;
                             end ;
                             escaping := FALSE ;
                             keystream := '' ;
                           end ;
              else
                begin { something else just clear it out }
                  escaping := FALSE ;
                  keystream := '' ;
                end ;
              end ;
            end ;
          end ;
        end
      else
        if Lo(ichar)=Ord(ESC) then
          begin { start to process ESC sequence }
            escaping := TRUE ;
            keystream := '' ;
          end
        else
          begin { just put character into event queue }
            with event do
              begin { may also need to fill in charcode better }
                what := evKeyDown ;
                charcode := Char(Lo(ichar)) ;
                if Lo(ichar)<$20 then
                  keycode := KEYCODES[ Lo(ichar) ]
                else
                  keycode := Lo(ichar) ;
              end ;
          end ;
      ichar := GetChar ;
    end ;
end ;

{----------------------------------------------------------
 Send ANSI code to set the current cursor position. Update
 the cposx/cposy variables.
----------------------------------------------------------}
Procedure TSRemote.SetCursor( ilin,icol : integer ) ;
begin
  PutLine( ESC+'['+Itos(ilin,1)+';'+Itos(icol,1)+'H' ) ;
  cposy := ilin ;
  cposx := icol ;
end ;

{----------------------------------------------------------
 Get the current video cursor position and update the
 serial screen output to match. Serial terminals use
 (1,1) as the upper left corner, TurboVision uses (0,0)
----------------------------------------------------------}
Procedure TSRemote.UpdateCursor ;
var
  cpos : TPoint ;
  modal : PView ;
  event : TEvent ;

begin
  modal := application^.TopView ; { point to top view }
  if Assigned(modal) then
    begin
      modal := PGroup(modal)^.current ; { point to focused view }
      if Assigned(modal) then
        begin
          cpos := modal^.cursor ;
          modal^.MakeGlobal( cpos,cpos ) ;
          if ( modal^.state and (sfCursorVis+sfFocused) )<>0 then
            if ( ( (cpos.x+1)<>cposx ) or ( (cpos.y+1)<>cposy ) ) then
              SetCursor( cpos.y+1,cpos.x+1 ) ;
        end ;
    end ;
end ;

{----------------------------------------------------------
 Send ANSI code for colour change
 Hi nibble = background colour
 Lo nibble = foregground colour
----------------------------------------------------------}
Procedure TSremote.SetColor( icol : byte ) ;
var
  prefix : string[3] ;
  fcol,bcol : byte ;
  oline : string[20] ;
begin
  fcol := icol and $0F ; { foreground colour }
  bcol := icol Shr 4 ;   { background colour }
  if (fcol>7) then
    prefix := '[1;'   { add bold for light colours }
  else
    prefix := '[0;' ; { no attributes for regular colours }
  fcol := fcol and $07 ;
  fcol := CMAP[fcol] ; { map to ANSI terminal colours }
  bcol := CMAP[bcol]+10 ;
  oline := ESC+prefix+Itos(fcol,1)+';'+Itos(bcol,1)+'m' ;
  PutLine( oline ) ;
end ;

{----------------------------------------------------------
 Basic transmit routine required from call in VIEWS.PAS.
 toptr   : pointer to destination of character movement
           ignore if this doesn't point to video memory
 fromptr : buffer holding character/attibute words to be
           sent to screen
 count   : number of words to move
----------------------------------------------------------}
Procedure TSRemote.Send( toptr,fromptr : pointer ; count : word ) ;
var
  ioff : word ;
  color : byte ;
  ichar : byte ;
  line,column : integer ;
  ibuffer : PDrawBuffer ;
  indx : integer ;
{.........................................................}
begin
  if (Ptr_rec(toptr).segment<>Ptr_rec(screenbuffer).segment) then
    Exit ; { not screen I/O }
  ibuffer := fromptr ;
  ioff := Ptr_rec(toptr).offset Div 2 ; { from words to characters }
  line := ioff Div COLS ;
  if (line<>0) then
    column := ioff Mod (line*COLS)
  else
    column := ioff ;
  SetCursor( line+1,column+1 ) ; { adjust for different origins }
  for indx := 0 to (count-1) do
    begin
      ichar := Lo( ibuffer^[indx] ) ;
      color := Hi( ibuffer^[indx] ) ;
      if (color<>scol) then
        begin
          scol := color ;
          SetColor( scol ) ;
        end ;
      if (ichar in [0..31]) then { adjust for lack of ANSI support for }
        ichar := ichar+$E0;      { IBM/PC graphic characters }
      PutChar( ichar ) ;
    end ;
end ;

{----------------------------------------------------------
 psb = 03 : 8 bits, no parity, 1 stop bit
 baud from RB_xxx constants above
----------------------------------------------------------}
Procedure TSRemote.InitPort( psb : byte ; baud : word ) ; assembler ;
asm
  les di,Self
  mov dx,es:[di].TSRemote.cport
  add dx,3       { dx = base+3 }
  mov al,$80
  out dx,al      { set DLAB bit to get at baud rates }
  sub dx,3       { dx = base }
  mov bx,baud    { set low baud byte }
  mov al,bl
  out dx,al
  add dx,1       { dx = base+1 }
  mov al,bh      { set high baud byte }
  out dx,al

  add dx,2       { dx = base+3 }
  mov al,psb     { parity/stop/bits }
  out dx,al

  sub dx,1       { dx = base+2 }
  mov al,$C7     { FIFO control register }
  out dx,al

  add dx,2       { dx = base+4 }
  mov al,$0B
  out dx,al
end ;

{----------------------------------------------------------
 Line status ah, Modem status al, same as Int14 ah=$03
----------------------------------------------------------}
Function TSRemote.GetStat : word ; assembler ;
asm
  les  di,Self
  mov  dx,es:[di].TSRemote.cport
  add  dx,5     { read LSR (base+5) }
  in   al,dx
  xchg ah,al    { put in ah }
  add  dx,1     { read MSR (base+6) }
  in   al,dx
end ;

{----------------------------------------------------------
 Get a character from the receiver buffer. Same as
 Int 14, ah=$02 - LSR in hi byte, char in lo byte
----------------------------------------------------------}
Function TSRemote.GetChar : word ; assembler ;
asm
  les  di,Self
  mov  dx,es:[di].TSRemote.cport
  add  dx,5        { read LSR first }
  in   al,dx
  xchg ah,al
  sub  dx,5
  in   al,dx       { read receiver buffer }
end ;

{----------------------------------------------------------
 Write obyte to transmitter register. Wait till safe to do
 so. This could all be interrupt driven.
----------------------------------------------------------}
Procedure TSRemote.PutChar( obyte : byte ) ; assembler ;
asm
  les  di,Self
  mov  dx,es:[di].TSRemote.cport
  add  dx,5       { get LSR }
@@1:
  in   al,dx
  test al,$20     { wait for empty transmitter register }
  jz   @@1
  sub  dx,5       { send byte }
  mov  al,obyte
  out  dx,al
end ;

{----------------------------------------------------------
 Transmit a string of characters
----------------------------------------------------------}
Procedure TSRemote.PutLine( oline : string ) ;
var
  indx : integer ;
begin
  for indx := 1 to Length( oline ) do
    PutChar( Ord(oline[indx]) ) ;
end ;

end .